//////////////////////////////////////////////////////////////////////
// UCSB - Electrical and Computer Engineering Dept.
//
// IOL_UART (IOL Uart Class)
//
// Version  When      Who                    Why
// -------  ----      ---                    ---
// 1.0      07/01/96  Daryl Fortney          Initial Design
//
// This class supports serial port access:
//  - DOS Direct (Standard UARTS)
//  - DOS Intellicon 8
//  - Windows
//  - UNIX
//
// Copyright 1996 UCSB.  All rights reserved.
//////////////////////////////////////////////////////////////////////
// Declaration
//////////////////////////////////////////////////////////////////////

#include <iostream>
using namespace std;

#ifndef IOL_UART_H
#define IOL_UART_H

#include "Iol.hpp"

// Event Timing
#ifndef EVENT_TIMING

	#ifdef IOL_UART_ALL
		#define IOL_UART_CHECKS
		#define IOL_UART_PRINTS
	#endif

	// Printing
	#define IOL_UART_PRINT_PRINT

	#ifdef  IOL_UART_PRINTS
		#define IOL_UART_PRINT_FLOWS
		#define IOL_UART_PRINT_ERRORS
	#endif

#endif

// UART Id
#define UART_PORTS	8
#define UART_1                  "/dev/ttyd1"    // Comm Port #1
#define UART_2                  "/dev/ttyd2"    // Comm Port #2
#define UART_3                  "/dev/ttyd3"    // Comm Port #3
#define UART_4                  "/dev/ttyd4"    // Comm Port #4
#define UART_5                  "/dev/ttyd5"    // Comm Port #5
#define UART_6                  "/dev/ttyd6"    // Comm Port #6
#define UART_7                  "/dev/ttyd7"    // Comm Port #7
#define UART_8                  "/dev/ttyd8"    // Comm Port #8
#define UART_ID                 NULL
extern char     UART_NAME[UART_PORTS][16];

// UART Mode
enum{   // Baud Rate
		UART_B                  =0xF000,        // Baud Rate Mask
		UART_300B               =0x1000,        // 300 bps
		UART_600B               =0x2000,        // 600 bps
		UART_1200B              =0x3000,        // 1200 bps
		UART_2400B              =0x4000,        // 2400 bps
		UART_4800B              =0x5000,        // 4800 bps
		UART_9600B              =0x6000,        // 9600 bps
		UART_19200B             =0x7000,        // 19200 bps
		UART_38400B             =0x8000,        // 38400 bps
		// Wordlength
		UART_W                  =0x0F00,        // Wordlength Mask
		UART_5W                 =0x0500,        // 5 Bits
		UART_6W                 =0x0600,        // 6 Bits
		UART_7W                 =0x0700,        // 7 Bits
		UART_8W                 =0x0800,        // 8 Bits
		// Parity
		UART_P                  =0x00F0,        // Parity Mask
		UART_NP                 =0x0010,        // No
		UART_OP                 =0x0020,        // Odd
		UART_EP                 =0x0030,        // Even
		UART_MP                 =0x0040,        // Mark
		UART_SP                 =0x0050,        // Space
		// Stopbits
		UART_S                  =0x000F,        // Stopbits Mask
		UART_1S                 =0x0001,        // 1
		UART_2S                 =0x0002,        // 2
};
#define UART_MODE               (UART_9600B|UART_8W|UART_NP|UART_1S)

#ifdef IOL_DOS_DIRECT
#define	UART_SIZE	2048	// FIFO Buffer Size

// Global RxTx UART Data
struct UART_DATA{
	void interrupt (far *OLD_ISR)(...);

	// UART Chip Address Info
	int     BASE;
	int     IRQ;
	int     IMR;

	// UART Chip Control Register
	volatile int    CONTROL;

	// Transmit FIFO
	volatile char   *TX;
	volatile int    TXH;
	volatile int    TXT;
	volatile int    TXL;

	// Receive FIFO
	volatile char   *RX;
	volatile int    RXH;
	volatile int    RXT;
	volatile int    RXL;
};

extern UART_DATA UD[4];
#endif

#ifdef IOL_WINDOWS
extern HANDLE  UD[4];
#endif

#ifdef IOL_UNIX
extern int     UD[4];
#endif

// Declaration
class IOL_UART{
	public:

	long    STATUS;

	char    ID[128];
	long    MODE;
	long    TYPE;
	long    PORT;
	long    TO;

	public:

	// Parameters
	enum{   
		TIMEOUT			=0x00,          // UART Timeout Period (1/10ths)
	};

	// Constants
	enum{ 
		// Types
		NONE                    =0x0000,        // No UART Detected
		T8250                   =0x0100,        // UART 8250 Detected
		T16450                  =0x0200,        // UART 1650 Detected
		T16550                  =0x0400,        // UART 16550 Detected
		T16550A                 =0x0800,        // UART 16550A Detected (FIFO)
		THIDDEN                 =0x1000,        // Hidden UART Detected
		// Registers
		IER                             =0x01,          // Interrupt Enable Register
		IIR                             =0x02,          // Interrupt Identification Register
		FCR                             =0x02,          // 16550 FIFO Control Register
		LCR                             =0x03,          // Line Control Register
		MCR                             =0x04,          // Modem Control Register
		LSR                             =0x05,          // Line Status Register
		MSR                             =0x06,          // Modem Status Register
		SCRATCH                 =0x07,          // Scratch Pad Register
		// Register IER Codes
		IER_RX                  =0x01,          // Enable Rx Data Available Intr.
		IER_TX                  =0x02,          // Enable Tx Hold Register Empty Intr.
		IER_LINE                =0x04,          // Enable Receiver Line Status Intr.
		IER_MODEM               =0x08,          // Enable Modem Status Intr.
		// Register IIR Codes
		IIR_NO                  =0x01,          // No Intr. Pending
		// Register FCR Codes
		FCR_DISABLE             =0x00,          // Disables FIFO
		FCR_ENABLE              =0x07,          // Enables FIFO & Resets
		// Register LSR Codes
		LSR_RX                  =0x01,          // Receiver Ready
		LSR_OE                  =0x02,          // Overrun Error
		LSR_PE                  =0x04,          // Parity Error
		LSR_FE                  =0x08,          // Framing Error
		LSR_BI                  =0x10,          // Break Interrupt
		LSR_TX                  =0x60,          // Transmitter Empty
		LSR_FIFO_ERR    =0x80,          // FE, PE, or BI in FIFO queue
	};

	// Constructor & Destructor
	IOL_UART(char *id=UART_ID,long mode=UART_MODE)
										{ STATUS=IOL_CLOSED; Open(id,mode); }
	~IOL_UART()                         { Close(); }

	// Maintainance Functions
	void    Print(void);

	// Initialization Functions
	BOOL    Open(char *id=UART_ID,long mode=UART_MODE);
	BOOL    Close(void);
	long    Detect(void);
	long    Status(void)                         { return STATUS; }

	// Attribute Functions
	char    *Id(void)                                       { return ID; }
	long    Mode(void)                                      { return MODE; }
	long    Mode(long mode);
	long    Parameter(long num);
	long    Parameter(long num,long value);
	long    Timeout()                                       { return Parameter(TIMEOUT); }
	long    Timeout(long value)                     { return Parameter(TIMEOUT,value); }

	// Tx & Rx Functions
	long    Tx(void *block=NULL,long num=1,long until=-1);
	long    Rx(void *block=NULL,long num=1,long until=-1);
#ifdef IOL_DOS_DIRECT
	void    TxFlush(void)                           { UD[PORT].TXH=UD[PORT].TXT=UD[PORT].TXL=0; }
	void    RxFlush(void)                           { UD[PORT].RXH=UD[PORT].RXT=UD[PORT].RXL=0; }
#endif
#ifdef IOL_DOS_CTI
	void	TxFlush(void)							{ }
	void	RxFlush(void);
#endif
#ifdef IOL_WINDOWS
	void    TxFlush(void)                           { }
	void    RxFlush(void)                           { }
#endif
	void    Flush(void)                             { TxFlush(); RxFlush(); }

#ifdef IOL_DOS_DIRECT
	// Direct UART Register Control
	char    Get(int r)                                      { return inp(UD[PORT].BASE+r); }
	void    Set(int r,char v)                       { outp(UD[PORT].BASE+r,v); }
	void    Or(int r,char v)                        { Set(r,v|Get(r)); }
	void    And(int r,char v)                       { Set(r,v&Get(r)); }
	void    Xor(int r,char v)                       { Set(r,v^Get(r)); }
#endif
};

#ifdef IOL_DOS_DIRECT
// ISR Prototypes
void far interrupt UART_ISR0();
void far interrupt UART_ISR1();
void far interrupt UART_ISR2();
void far interrupt UART_ISR3();
void UART_HANDLER(const int port);
extern void far *UART_ISR[4];
#endif

#endif
