/*	Hapti Guidance
*	Author: Andreas Nygaard, 2008
*	
*	- Position position control scheme
*	- Guidance: slave based (xd-xs)
*	- Omni force control: PD, input: Omni cart. pos - Aesop cart. pos, output: cart. force 
*	- AESOP velocity control: PD, input: Omni cart. pos - Aesop cart. pos - guidance vel, output: Aesop cart. velocity
*/

#if defined(WIN32)
# include <conio.h>
#else
# include "conio.h"
#endif

#include <string>
#include <HD/hd.h>
#include <HDU/hduError.h>
#include <HDU/hduVector.h>
#include <HDU/hduMatrix.h>
#include <process.h>
#include <windows.h>
#include <stdio.h>
#include <iostream>
#include <fstream>
#include <math.h>
#include <ctime>

#include "Aesop/a3000.h"

#define dirX 0
#define	dirY 1
#define dirZ 2

//WriteSysParam alternatives
#define HD_POSITION 0
#define HD_VELOCITY 1
#define HD_FORCE 2
#define HD_BUTTON_1 3
#define R_POSITION 4
#define R_INIT_POSITION 5
#define R_VELOCITY 6

//Guidance alternatives
#define NONE_GUIDANCE 0
#define PLANE_GUIDANCE 1
#define LINE_GUIDANCE 2
#define POINT_GUIDANCE 3

// System variables
struct DeviceDisplayState
{
	int writeState;
	HDdouble position[3];	
	HDdouble HinitPos[3];
	HDdouble velocity[3]; 
	HDdouble force[3];
	HDdouble Rposition[3];
	HDdouble RinitPosition[3];
	HDdouble Rvelocity[3];
	float RjointVel[4];
	float RjointAng[7];
	float RposMeasure[6];
	HDboolean button1State;
};

//Parameters for Guidance
struct GuidanceParameter
{
	int guidanceType;
	int LastGuidanceType;
	hduVector3D <HDdouble> lineVector;
	hduVector3D <HDdouble> vdLast;
	hduVector3D <HDdouble> vd;
	hduVector3D <HDdouble> PerrLast;
	hduVector3D <HDdouble> Rpos;
	hduVector3D <HDdouble> Hpos;
	hduVector3D <HDdouble> p1;
	hduVector3D <HDdouble> p2;
	hduVector3D <HDdouble> p3;
	HDdouble MaxForce;
	hduVector3D <HDdouble> Km;
	hduVector3D <HDdouble> Cm;
	hduVector3D <HDdouble> Ks;
	hduVector3D <HDdouble> Cs;
	hduVector3D <HDdouble> Kg;
	hduVector3D <HDdouble> Cg;
	HDdouble MaxDistance;
	double TimeInit, TimeLast, TimeNow, TimeFin;

};

//*** Robot Parameters ***
int DS_PORT = 3;
const double L2= 384.4879;
const double L31= 76.7056;
const double L32= 304.8001;
double L3, qe;

//*** Global variables ***
DeviceDisplayState sysParam; //Secure struct object, use WriteSysParam / ReadSysParam to access
GuidanceParameter guideParam;
HDSchedulerHandle gCallbackHandle = 0;
double tini, tfin, tr, lastTr; //times
boolean run;
boolean first;
boolean button;
boolean Ron;
int testNum =0;
int sign=1;

//*** ReadKeyHit parameters ***
int planeCount;
boolean writeFirst;
boolean lineOK;
boolean planeOK;
boolean planeNext;
boolean buttonLast;
boolean CheckGuidParam;
hduVector3D <HDdouble> point1;
hduVector3D <HDdouble> point2;

//*** Functions init ***
void mainloop(void *pUserData);
void CalculateForce(void *pUserData);
void AesopCartToJointVel(void *pUserData);
void AesopForwardKinematics(void *pUserData);
void writeFile(void *pUserData);
void readKeyHit(void *pUserData);
HDCallbackCode HDCALLBACK HapticLoop(void *pUserData);
HDCallbackCode HDCALLBACK WriteSysParam(void *pUserData);
HDCallbackCode HDCALLBACK ReadSysParam(void *pUserData);




/*	Main
*	- Init AESOP/Omni and global parameters
*	- Stop AESOP/Omni after mainloop is quit
*/
int main()
{
	sysParam.force[dirX]=0;
	sysParam.force[dirY]=0;
	sysParam.force[dirZ]=0;

	//Default Guidance parameters
	guideParam.guidanceType=NONE_GUIDANCE;
	guideParam.LastGuidanceType=guideParam.guidanceType;
	guideParam.MaxDistance=10;
	guideParam.MaxForce=3;
	guideParam.lineVector.set(0,0,0);
	guideParam.vdLast.set(0,0,0);
	guideParam.PerrLast.set(0,0,0);
	guideParam.p1.set(0,0,0);
	guideParam.p2.set(0,0,0);
	guideParam.p3.set(0,0,0);

	//Master spring/damper gain K/C
	guideParam.Km.set(.01, .01, .0);
	guideParam.Cm.set(.001, .001, .005);

	//Slave spring/damper gain K/C
	guideParam.Ks.set(1, 1, 1.5);
	guideParam.Cs.set(1, 1, .8);

	//Guidance spring/damper gain K/C
	guideParam.Kg.set(.5, .5, .5);
	guideParam.Cg.set(.005, .005, .001);

	run = true;
	first=true;
	Ron=true;
	button=false;
	buttonLast=button;
	lineOK=false;
	planeOK=false;
	planeNext=true;
	CheckGuidParam=false;
	writeFirst=true;
	point1.set(0,0,0);
	point2.set(0,0,0);
	planeCount = 1;

	//create struct object to be used all but for Omni manipulations
	DeviceDisplayState temp;
	temp.button1State=0;
	temp.Rvelocity[0]=0;
	temp.Rvelocity[1]=0;
	temp.Rvelocity[2]=0;

	//*** Init Robot ***
	if (!Initialize(DS_PORT)){
		cout << "Working to open AESOP Communications.\n\n" << endl;
		while(!Initialize(DS_PORT)){for(int i=0;i<1000000000;i++){}}
		//exit(1);
	}
	JtPosReq(temp.RjointAng);
	qe=temp.RjointAng[6];
	L3=L31+L32*cos(qe);

	//*** Init Haptic ***
	HHD hHD = hdInitDevice(HD_DEFAULT_DEVICE);

	gCallbackHandle = hdScheduleAsynchronous(HapticLoop, &sysParam, HD_MAX_SCHEDULER_PRIORITY);
    
	hdEnable(HD_FORCE_OUTPUT);  // Enable forces
	hdStartScheduler();// Start the servo scheduler
	
	printf("***************************************** \n");
	printf("*                                       * \n");
	printf("*              Welcome to               * \n");
	printf("*                                       * \n");
	printf("*        Haptic Guidance System         * \n");
	printf("*                                       * \n");
	printf("*                                       * \n");
	printf("*            Andreas Nygaard            * \n");
	printf("*    NTNU / The Interventional Center   * \n");
	printf("*                 2008                  * \n");
	printf("*                                       * \n");
	printf("***************************************** \n");
	printf("\n\nInstructions:\n0   - No Guidance\n1   - Plane Guidance\n");
	printf("2   - Line Guidance\n3   - Point Guidance\n+/- - Increase / Decrease Max Distance\n");
	printf("\nPress Q to Quit!\n");
	
	//*** Mainloop runs until stopped ***
	mainloop(&temp);

	//Stop robot
	temp.RjointVel[0]=0;
	temp.RjointVel[1]=0;
	temp.RjointVel[2]=0;
	temp.RjointVel[3]=0;
	MoveAllJointsContinous(temp.RjointVel);

	//Stop the haptic
	hdStopScheduler();
    hdUnschedule(gCallbackHandle);
    hdDisableDevice(hHD);
	
	return 0;

}

/*	Mainloop runs until Q is hit
*	- updates AESOP and Omni parameters, runs guidance as long as button on Omni is pushed 
*	Control buttons
*	Q - Quit
*	0 - NONE_GUIDANCE
*	1 - PLANE_GUIDANCE
*	2 - LINE_GUIDANCE
*	3 - POINT_GUIDANCE
*	+/- Increase/decrease max distance
*	o - Switches AESOP ON/OFF
*/
void mainloop(void *pUserData)
{
	DeviceDisplayState *temp = static_cast<DeviceDisplayState *>(pUserData);
	guideParam.TimeInit = GetTickCount();
	tini = guideParam.TimeInit;
	
	while(run)
	{
		//Read values from haptic
		hdScheduleSynchronous(ReadSysParam, temp, HD_DEFAULT_SCHEDULER_PRIORITY);
		button=temp->button1State;
		
		//Read values from robot
		JtPosReq(temp->RjointAng);
		temp->RjointAng[0]=temp->RjointAng[0]*(float)25.4;

		AesopForwardKinematics(temp); //convert joint to position space
	
		readKeyHit(temp); // checks for keybord hits

		if(button)
		{
			//printf("%f %f %f %f %f %f \n", temp->velocity[0], temp->velocity[1], temp->velocity[2], temp->RjointVel[0], temp->RjointVel[1], temp->RjointVel[2]);
			
			if(first)
			{
				//printf("ACTIVE!\n");
				//Init pos of robot
				memcpy(&temp->RinitPosition, &temp->Rposition, sizeof(temp->Rposition));
				//Init pos of haptic
				memcpy(&temp->HinitPos, &temp->position, sizeof(temp->position));
				//sign=sign*(-1);
				writeFile(temp);
				first=false;
			}
			else if(!first)
			{
				CalculateForce(temp);
				AesopCartToJointVel(temp);
				if(Ron)MoveAllJointsContinous(temp->RjointVel);
			}
			writeFile(temp);
		}
		else if(!button && buttonLast)
		{
			// Give zero force to haptic
			temp->force[0]=0;
			temp->force[1]=0;
			temp->force[2]=0;
			temp->writeState=HD_FORCE;
			WriteSysParam(temp);

			// Stop robot
			temp->RjointVel[0]=0;
			temp->RjointVel[1]=0;
			temp->RjointVel[2]=0;
			temp->RjointVel[3]=0;
			MoveAllJointsContinous(temp->RjointVel);

			first=true;
		}
		buttonLast=button;
		guideParam.LastGuidanceType=guideParam.guidanceType;
	}
}


/*	Calculates force to put onto the Omni
*	- calculations in AESOP coordinate space
*	- output in omni coordinate space
*/
void CalculateForce(void *pUserData)
{
	DeviceDisplayState *temp = static_cast<DeviceDisplayState *>(pUserData);

	double distGain;
	double maxForce = 3.3;//Newton
	int GuidanceType = guideParam.guidanceType;

	HDdouble vlM, A, B, C, D, d;
	hduVector3D <HDdouble> forceTemp;
	hduVector3D <HDdouble> forceLast;
	hduVector3D <HDdouble> VelTemp;
	hduVector3D <HDdouble> Rvel;
	hduVector3D <HDdouble> Rpos;
	hduVector3D <HDdouble> Hpos;
	hduVector3D <HDdouble> RposInit;
	hduVector3D <HDdouble> HposInit;
	hduVector3D <HDdouble> Perr;
	hduVector3D <HDdouble> dPerr;
	hduVector3D <HDdouble> vr;
	hduVector3D <HDdouble> v;
	hduVector3D <HDdouble> vd;
	hduVector3D <HDdouble> dvd; //differentiated vd
	hduVector3D <HDdouble> pd;
	hduVector3D <HDdouble> pi;
	hduVector3D <HDdouble> pr;
	hduVector3D <HDdouble> ph;
	hduVector3D <HDdouble> p1;
	hduVector3D <HDdouble> p2;
	hduVector3D <HDdouble> p3;
	hduVector3D <HDdouble> v1;
	hduVector3D <HDdouble> v2;
	hduVector3D <HDdouble> n;
	hduVector3D <HDdouble> nn;

	//Update time
	guideParam.TimeFin = GetTickCount();
	guideParam.TimeNow = (guideParam.TimeFin - guideParam.TimeInit)/1000;
	
	//Set pos vectors
	Rpos.set(temp->Rposition[0], temp->Rposition[1], temp->Rposition[2]);
	RposInit.set(temp->RinitPosition[0], temp->RinitPosition[1], temp->RinitPosition[2]);
	Hpos.set(temp->position[2], temp->position[0], temp->position[1]);		//convert to aesop coord.
	HposInit.set(temp->HinitPos[2], temp->HinitPos[0], temp->HinitPos[1]);	//convert to aesop coord.
	
	//Convert to relative positions
	Rpos= Rpos - RposInit;
	Hpos= Hpos - HposInit;

	p1.set(0,0,0);
	p2.set(0,0,0);
	p3.set(0,0,0);
	VelTemp.set(0,0,0);
	forceTemp.set(0,0,0);
	pi.set(0,0,0);
	vd.set(0,0,0);
	Rvel.set(0,0,0);
	forceLast.set(temp->force[0], temp->force[1], temp->force[2]);
		
	//**** Calculates deviation vector vd ****
	switch(GuidanceType)
	{
		case 0: // NONE
		{
			break;
		}
		case 1: // Plane guidance
		{
			if(planeOK)
			{
				p1=guideParam.p1;
				p2=guideParam.p2;
				p3=guideParam.p3;
				v1=p1-p2;
				v2=p1-p3;
				n=crossProduct(v1,v2);
				nn=normalize(n);
				A=n[0];
				B=n[1];
				C=n[2];
				D= -(n[0]*v1[0]+n[1]*v1[1]+n[2]*v1[2]);
				d= (A*Rpos[0]+B*Rpos[1]+C*Rpos[2]+D)/sqrt(A*A+B*B+C*C);
				vd=-nn*d;

				for(int i=0;i<3;i++)
				{
					if(vd[i]<0)vd[i]=0;
				}
			}
			break;
		}
		case 2: // Line guidance
		{
			if(lineOK)
			{
				v = guideParam.lineVector;
				// Calculate deviation vd from vectorline v
				//v.set(1,1,0);
				v.normalize();
				vr = Rpos - pi;
				vlM = dotProduct(v,vr) / v.magnitude();
				pd = pi + vlM * v;
				vd = pd - Rpos;
			}
			break;
		}
		case 3: // Point guidance
		{
			vd = pi-Rpos;
			
			break;	
		}
		default:
		{
			printf("No guidance type chosen!/n");
			break;
		}
	}
	
	guideParam.vd=vd;

	//**** Calculates Pos error and diff Pos error ****
	Perr = Hpos - Rpos;
	dPerr = (Perr - guideParam.PerrLast) / (guideParam.TimeNow - guideParam.TimeLast);
	dvd = (vd - guideParam.vdLast) / (guideParam.TimeNow - guideParam.TimeLast);

	//*** Max Distance scaling
	distGain=vd.magnitude()/guideParam.MaxDistance;
	if(distGain<-1)distGain=-1;
	else if(distGain>1)distGain=1;

	//**** Calculates Omni Force ****
	forceTemp = -(Perr*guideParam.Km + dPerr*guideParam.Cm) +(vd*distGain*guideParam.Kg + dvd*guideParam.Cg);
	for(int i=0;i<3;i++)
	{
		//limits force output
		if(forceTemp[i]>3)
		{
			forceTemp[i]=3;
		}
		else if(forceTemp[i]<-3)
		{
			forceTemp[i]=-3;
		}
	}
	// Transfer to Omni coordinate space
	temp->force[0] = forceTemp[1];
	temp->force[1] = forceTemp[2];
	temp->force[2] = forceTemp[0];


	//**** Calculate Aesop Velocity ****
	VelTemp = Perr*guideParam.Ks + dPerr*guideParam.Cs; 
	temp->Rvelocity[0] = VelTemp[0];
	temp->Rvelocity[1] = VelTemp[1];
	temp->Rvelocity[2] = VelTemp[2];


	// Filter and Set force onto Omni
	for(int i=1; i<3; i++)
	{
		double limit = .3;
		if(forceTemp[i]-forceLast[i]<-limit)forceTemp[i]=forceLast[i]-limit;
		else if(forceTemp[i]-forceLast[i]>limit)forceTemp[i]=forceLast[i]+limit;
	}
	hdScheduleSynchronous(WriteSysParam, temp, HD_DEFAULT_SCHEDULER_PRIORITY);
	
	//For log purposes
	guideParam.Rpos = Rpos;
	guideParam.Hpos = Hpos;
	
	//Update dvd parameters
	memcpy(&guideParam.PerrLast, &Perr, sizeof(Perr));
	memcpy(&guideParam.vdLast, &vd, sizeof(vd));
	memcpy(&guideParam.TimeLast, &guideParam.TimeNow, sizeof(guideParam.TimeNow));
}


/** All calculations in AESOP coordinate space, 
*   Joint space: q1[mm]/q2[rad]/q3[rad], dq1[mm/s]/dq2[rad/s]/dq3[rad/s]
*	Cartesian space:  dX[mm/s]
*/
void AesopCartToJointVel(void *pUserData)
{
	DeviceDisplayState *temp = static_cast<DeviceDisplayState *>(pUserData);

	//Parameters of the AESOP, joint angles, arm lengths, jacobian elements
	HDdouble q1, q2, q3, dq1, dq2, dq3, dx, dy, dz, invDetJ;
	HDdouble j11, j12, j13, j21, j22, j23, j31, j32, j33;
	HDdouble c11, c12, c13, c21, c22, c23, c31, c32, c33;
	
	q1= (HDdouble) temp->RjointAng[0];
	q2= (HDdouble) temp->RjointAng[1];
	q3= (HDdouble) temp->RjointAng[2];
	
	double g[3];
	g[0]=1;
	g[1]=1;
	g[2]=1;
	dx= temp->Rvelocity[dirX]*g[0];
	dy=	temp->Rvelocity[dirY]*g[1];
	dz= temp->Rvelocity[dirZ]*g[2];
	
	//Jacobian
	j11=0;
	j12=-L2*sin(q2)-L3*sin(q2+q3);
	j13=-L3*sin(q2+q3);
	j21=0;
	j22=L3*cos(q2+q3)+L2*cos(q2);
	j23=L3*cos(q2+q3);
	j31=1;
	j32=0;
	j33=0;

	//****Calculate Inverse of Jacobian***
	//Determinant of Jacobian
	invDetJ=1/((j11*j22*j33 + j12*j23*j31 + j13*j21*j32)-(j31*j22*j13 + j32*j23*j11 + j33*j21*j12));

	//Cofactor elements of Jacobian
	c11=  (j22*j33 - j23*j32);
	c12= -(j21*j33 - j23*j31);
	c13=  (j21*j32 - j22*j31);
	c21= -(j12*j33 - j13*j32);
	c22=  (j11*j33 - j13*j31);
	c23= -(j11*j32 - j12*j31);
	c31=  (j12*j23 - j13*j22);
	c32= -(j11*j23 - j13*j21);
	c33=  (j11*j22 - j12*j21);


	//**** dTheta = invJ*dX ****
	dq1= invDetJ*(c11*dx + c21*dy + c31*dz);
	dq2= invDetJ*(c12*dx + c22*dy + c32*dz);
	dq3= invDetJ*(c13*dx + c23*dy + c33*dz);
	//printf("%f %f %f %f %f %f %f %f %f\n", q1, q2, q3, dq1, dq2, dq3, dx, dy, dz);
	//Check velocity limits
	if(dq1>30)dq1=30;
	else if(dq1<-30)dq1=-30;

	if(dq2>0.3)dq2=0.3;
	else if(dq2<-0.3)dq2=-0.3;

	if(dq3>1)dq3=1;
	else if(dq3<-1)dq3=-1;

	//convert from mm/s to inch/s or something
	dq1=dq1/2.54; 

	temp->RjointVel[0]=(float)dq1;
	temp->RjointVel[1]=(float)dq2;
	temp->RjointVel[2]=(float)dq3;
	temp->RjointVel[3]= 0;

}


/*	Calculates Cartesian position of the AESOP
*	based on its joint angles
*	- result is given in AESOP coordinate frame
*/
void AesopForwardKinematics(void *pUserData)
{
	DeviceDisplayState *temp = static_cast<DeviceDisplayState *>(pUserData);
	double q1, q2, q3, x, y, z;
	
	//Asign values
	q1=temp->RjointAng[0];
	q2=temp->RjointAng[1];
	q3=temp->RjointAng[2];

	//Forward kinematics
	x= L2*cos(q2) + L3*cos(q2+q3);
	y= L2*sin(q2) + L3*sin(q2+q3);
	z= L32*sin(qe) + q1;
	
	//Put transformed values back
	temp->Rposition[dirX]= x;
	temp->Rposition[dirY]= y;
	temp->Rposition[dirZ]= z;

}


/*	ReadKeyHit
*	- reads when keybord buttons are pushed 
*	Control buttons
*	Q - Quit
*	0 - NONE_GUIDANCE
*	1 - PLANE_GUIDANCE
*	2 - LINE_GUIDANCE
*	3 - POINT_GUIDANCE
*	+/- Increase/decrease max distance
*	o - Switches AESOP ON/OFF
*/
void readKeyHit(void *pUserData)
{
	DeviceDisplayState *temp = static_cast<DeviceDisplayState *>(pUserData);
	button=temp->button1State;
	int keypress;
	int b=0; //Used for changing axis when tuning gains
	
	// Check if buttons are pushed
	if(_kbhit())
	{
		keypress = _getch();
		if(keypress == 'Q'||keypress =='q')
		{
			run = false;
		}
		else if(keypress == '0')
		{
			printf("\nGuidance type: None\n");	
			guideParam.guidanceType=NONE_GUIDANCE;
		}
		else if(keypress == '1')
		{
			printf("\nGuidance type: Plane\n");
			guideParam.guidanceType=PLANE_GUIDANCE;
		}
		else if(keypress == '2')
		{
			printf("\nGuidance type: Line\n");
			guideParam.guidanceType=LINE_GUIDANCE;
		}
		else if(keypress == '3')
		{
			printf("\nGuidance type: Point\n");	
			guideParam.guidanceType=POINT_GUIDANCE;
		}
		else if(keypress == '+')
		{
			guideParam.MaxDistance++;
			printf("MaxDistance: %f\n",guideParam.MaxDistance);	
		}
		else if(keypress == '-')
		{
			guideParam.MaxDistance--;
			if(guideParam.MaxDistance<0)guideParam.MaxDistance=0;
			printf("MaxDistance: %f\n",guideParam.MaxDistance);	

				
		}
		else if(keypress == '9')
		{
			guideParam.Km[b] = guideParam.Km[b] + .01;
			printf("Km[%i]: %f\n",b ,guideParam.Km[b]);	
		}
		else if(keypress == '6')
		{
			guideParam.Km[b] = guideParam.Km[b] - .01;
			if(guideParam.Km[b]<0)guideParam.Km[b]=0;
			printf("Km[%i]: %f\n",b ,guideParam.Km[b]);	
		}
		else if(keypress == 'o')
		{
			if(Ron)
			{
				Ron=false;
				printf("Robot: OFF\n");
			}
			else 
			{
				Ron=true;
				printf("Robot: ON\n");
			}
		}
	}
	// For creating the line with which to guide along
	if(guideParam.guidanceType!=guideParam.LastGuidanceType)CheckGuidParam=true;

	if(CheckGuidParam)
	{
		if(guideParam.guidanceType==LINE_GUIDANCE)
		{	
			planeOK=false;
			if(writeFirst)
			{
				printf("\nPush button and drag haptic device between \ntwo end points creating a line! \n");
				writeFirst=false;
				guideParam.lineVector.set(0,0,0);
			}
			if(button && !buttonLast)
			{
				point1.set(temp->Rposition[0], temp->Rposition[1], temp->Rposition[2]);
				printf("1. OK!\n");
			
			}
			if(!button && buttonLast)
			{
				point2.set(temp->Rposition[0], temp->Rposition[1], temp->Rposition[2]);
				lineOK=true;
				CheckGuidParam=false;
				writeFirst=true;
				guideParam.lineVector=point2-point1;
				guideParam.lineVector.normalize();
				printf("2. OK, line is set!\n");
			}
		}
		else if(guideParam.guidanceType==PLANE_GUIDANCE)
		{
			lineOK=false;
			if(writeFirst)
			{
				printf("\nWe're now going to creat a plane out of \nthree points.\n");
				printf("\nPush button and drag haptic device to a \npoint and then let go of the button! \n");
				writeFirst=false;
				guideParam.p1.set(0,0,0);
				guideParam.p2.set(0,0,0);
				guideParam.p3.set(0,0,0);
			}
			if(button && !buttonLast)
			{
				planeNext=true;
			}
			if(!button && buttonLast && planeNext)
			{
				if(planeCount==1)
				{
					guideParam.p1.set(temp->Rposition[0], temp->Rposition[1], temp->Rposition[2]);
					printf("1. OK, next point! \n");
					planeCount++;
					planeNext=false;
				}
				else if(planeCount==2)
				{
					guideParam.p2.set(temp->Rposition[0], temp->Rposition[1], temp->Rposition[2]);
					printf("2. OK, and the last point! \n");
					planeCount++;
					planeNext=false;
				}
				else if(planeCount==3)
				{
					guideParam.p3.set(temp->Rposition[0], temp->Rposition[1], temp->Rposition[2]);
					printf("3. Thanks, the plane is set! \n");
					planeCount=1;
					planeOK=true;
					CheckGuidParam=false;
					writeFirst=true;
					planeNext=true;
					
					guideParam.p3=guideParam.p3-guideParam.p1;
					guideParam.p2=guideParam.p2-guideParam.p1;
					guideParam.p1=guideParam.p1-guideParam.p1;					
					
					/*guideParam.p2.set(guideParam.p2[0]-guideParam.p1[0], guideParam.p2[1]-guideParam.p1[1], guideParam.p2[2]-guideParam.p1[2]);
					guideParam.p3.set(guideParam.p3[0]-guideParam.p1[0], guideParam.p3[1]-guideParam.p1[1], guideParam.p3[2]-guideParam.p1[2]);
					guideParam.p1.set(guideParam.p1[0]-guideParam.p1[0], guideParam.p1[1]-guideParam.p1[1], guideParam.p1[2]-guideParam.p1[2]);
					*/
					/*printf("p1: %f %f %f\n", guideParam.p1[0], guideParam.p1[1], guideParam.p1[2]);
					printf("p2: %f %f %f\n", guideParam.p2[0], guideParam.p2[1], guideParam.p2[2]);
					printf("p3: %f %f %f\n", guideParam.p3[0], guideParam.p3[1], guideParam.p3[2]);
					*/
					CheckGuidParam=false;
				}
			}
		}
		else if(guideParam.guidanceType==NONE_GUIDANCE)
		{
			CheckGuidParam=false;
			planeOK=false;
			lineOK=false;
		}
		else if(guideParam.guidanceType==POINT_GUIDANCE)
		{
			CheckGuidParam=false;
			planeOK=false;
			lineOK=false;
		}
	}
}


/*	For log/debug purposes
*	- writes central variables to log file at Results/res%i.m (%i is number of test)
*/
void writeFile(void *pUserData)
{
	DeviceDisplayState *temp = static_cast<DeviceDisplayState *>(pUserData);

	if(first)
	{
		tini= GetTickCount();
		testNum=testNum+1;

		char ResFile[200];
		sprintf(ResFile, "Results/res%i.m",testNum);
		FILE *stream1;
		stream1 = fopen(ResFile,"a");
		fprintf(stream1,"Guidance type: %i \n",guideParam.guidanceType);
		if(guideParam.lineVector[0]!=0)fprintf(stream1,"Line Vector: %f %f %f \n",guideParam.lineVector[0],guideParam.lineVector[1],guideParam.lineVector[2]);
		if(guideParam.p2[0]!=0)
		{
			fprintf(stream1,"Point 1: %f %f %f \n",guideParam.p1[0],guideParam.p1[1],guideParam.p1[2]);
			fprintf(stream1,"Point 2: %f %f %f \n",guideParam.p2[0],guideParam.p2[1],guideParam.p2[2]);
			fprintf(stream1,"Point 3: %f %f %f \n",guideParam.p3[0],guideParam.p3[1],guideParam.p3[2]);
		}
		fprintf(stream1,"\nMaster K: %f %f %f C: %f %f %f \n",guideParam.Km[0],guideParam.Km[1],guideParam.Km[2],guideParam.Cm[0],guideParam.Cm[1],guideParam.Cm[2]);
		fprintf(stream1,"Slave  K: %f %f %f C: %f %f %f \n",guideParam.Ks[0],guideParam.Ks[1],guideParam.Ks[2],guideParam.Cs[0],guideParam.Cs[1],guideParam.Cs[2]);
		fprintf(stream1,"Guide  K: %f %f %f C: %f %f %f \n",guideParam.Kg[0],guideParam.Kg[1],guideParam.Kg[2],guideParam.Cg[0],guideParam.Cg[1],guideParam.Cg[2]);
		fprintf(stream1,"Max Distance: %f \n",guideParam.MaxDistance);
		fprintf(stream1,"\nContent: tr, Rpos, Hpos, Hforce, vd\n\n");
		//fprintf(stream1,"Guidance type: %i \nContent: tr, temp->RjointAng[0], temp->RjointAng[1], temp->RjointAng[2], temp->RjointVel[0], temp->RjointVel[1], temp->RjointVel[2], temp->Rposition[0], temp->Rposition[1], temp->Rposition[2], temp->velocity[dirZ], -temp->velocity[dirX], temp->velocity[dirY] \n\n", guideParam.guidanceType);	
		fclose(stream1);
		first=false;
	}

	//Start time at zero
	tfin = GetTickCount();
	tr=(tfin-tini)/1000;
	lastTr=tr;

	char ResFile[200];
	sprintf(ResFile, "Results/res%i.m",testNum);
	FILE *stream1;
	stream1 = fopen(ResFile,"a");
	//fprintf(stream1,"%f %f %f %f \n",tr, temp->force[0], temp->force[1], temp->force[2]);
	fprintf(stream1,"%f %f %f %f %f %f %f %f %f %f %f %f %f\n",tr, guideParam.Rpos[0], guideParam.Rpos[1], guideParam.Rpos[2], guideParam.Hpos[0], guideParam.Hpos[1], guideParam.Hpos[2], temp->force[2], temp->force[0], temp->force[1], guideParam.vd[0], guideParam.vd[1], guideParam.vd[2]);
		
	fclose(stream1);

}

/*	HapticLoop
*	- Runs as separate thread
*	- Read status of Omni
*	- Set forces to Omni
*/
HDCallbackCode HDCALLBACK HapticLoop(void *pUserData)
{
	//Variables
	DeviceDisplayState *SysParamTemp = static_cast<DeviceDisplayState *>(pUserData);
	int nButtons=0;
	HDdouble jointAng[3];

	HHD hHD = hdGetCurrentDevice();
	hdBeginFrame(hHD);

	//Read values from haptic
	hdGetDoublev (HD_CURRENT_JOINT_ANGLES, (HDdouble *)&jointAng);
	hdGetDoublev (HD_CURRENT_VELOCITY, SysParamTemp->velocity);
	hdGetDoublev (HD_CURRENT_POSITION, SysParamTemp->position);
	hdGetIntegerv (HD_CURRENT_BUTTONS, &nButtons);
	SysParamTemp->button1State = (nButtons & HD_DEVICE_BUTTON_1) ? TRUE : FALSE;
	
	//Set force to haptic
	hdSetDoublev(HD_CURRENT_FORCE, SysParamTemp->force);

	hdEndFrame(hHD);

	return HD_CALLBACK_CONTINUE;
}

/*	ReadSysParameters
*	- Synchronised access of Omni parameters
*	- Use this function to read the variables of the Omni
*/
HDCallbackCode HDCALLBACK ReadSysParam(void *pUserData)
{
	DeviceDisplayState *SysParamTemp = static_cast<DeviceDisplayState *>(pUserData);
	
	memcpy(&SysParamTemp->force, &sysParam.force, sizeof(sysParam.force));
	memcpy(&SysParamTemp->position, &sysParam.position, sizeof(sysParam.position));
	memcpy(&SysParamTemp->velocity, &sysParam.velocity, sizeof(sysParam.velocity));
	memcpy(&SysParamTemp->button1State, &sysParam.button1State, sizeof(sysParam.button1State));

	return HD_CALLBACK_DONE;
}

/*	WriteSysParameters
*	- Synchronised access of Omni parameters
*	- Use this function to set the variables of the Omni
*/
HDCallbackCode HDCALLBACK WriteSysParam(void *pUserData)
{
	DeviceDisplayState* SysParamTemp= (DeviceDisplayState *) pUserData;
	SysParamTemp->writeState=HD_FORCE;
	switch(SysParamTemp->writeState)
	{
		case 0: // Writes to private HDposition variable
		{
			memcpy(&sysParam.position,&SysParamTemp->position, sizeof(SysParamTemp->position));
	
			break;
		}
		case 1: // Writes to private HDvelocity variable
		{
			memcpy(&sysParam.velocity,&SysParamTemp->velocity, sizeof(SysParamTemp->velocity));

			break;
		}
		case 2: // Writes to private HDforce variable
		{
			memcpy(&sysParam.force,&SysParamTemp->force, sizeof(SysParamTemp->force));
			
			break;
		}
		case 3: // Writes to private HDbutton1State variable
		{
			memcpy(&sysParam.button1State,&SysParamTemp->button1State, sizeof(SysParamTemp->button1State));

			break;
		}
		default:
		{
			break;
		}
	}

	return HD_CALLBACK_DONE;
}



